import os.path
import re
from subprocess import Popen, PIPE
import pytest


@pytest.fixture()
def clipboard():
    """Retrieve clipboard content."""
    def callback(raw: bool=False) -> str:
        process = Popen(['osascript', '-e', u'the clipboard as \xabclass HTML\xbb'], stdout=PIPE)
        text = process.communicate()[0].decode('utf-8')
        if raw:
            return text
        hex_text = re.match(r'\xabdata HTML(.*)\xbb', text).group(1)
        return bytes.fromhex(hex_text).decode('utf-8')
    return callback


@pytest.fixture()
def resource():
    """Retrieve path to test resource file."""
    resource_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources')

    def callback(*paths: str) -> str:
        return os.path.join(resource_dir, *paths)
    return callback



import pygclip.colors as c


def test_darken_color():
    """Test a few examples of darkening web colors."""
    assert c.darken_color('#000000', 0.5) == '#000000'
    assert c.darken_color('#ffffff', 0.5) == '#7f7f7f'
    assert c.darken_color('#ffffff', 0.8) == '#323232'



"""
    pygclip.cli
    ~~~~~~~~~~~

    A simple command line application to run pygclip.

    :copyright: © 2018 by Bryan Lee McKelvey.
    :license: MIT, see LICENSE for more details.
"""

import argparse
import logging
import os.path
import sys
from typing import Union

from pygclip.html import generate_html
from pygclip.io import write_html_to_clipboard
from pygclip.logging import setup_logging


def _process_arguments(
    lexer: Union[str, None],
    style: Union[str, None],
    path: Union[str, None],
    clipboard: bool,
) -> None:
    html = generate_html(lexer=lexer, style=style, path=path, clipboard=clipboard)
    write_html_to_clipboard(html)


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Run pygclip.')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true', help='debug output')
    parser.add_argument('-l', '--lexer', metavar='<lexer>', dest='lexer', help='lexer name')
    parser.add_argument('-s', '--style', metavar='<style>', dest='style', help='style name')
    parser.add_argument('-c', '--clipboard', dest='clipboard', action='store_true', help='read from clipboard')
    parser.add_argument('path', nargs='?', help='input file path')
    args = parser.parse_args(argv)

    setup_logging(args.debug)

    if args.path is not None and not os.path.isfile(args.path):
        raise FileNotFoundError('File does not exist: {}'.format(args.path))

    _process_arguments(
        lexer=args.lexer,
        style=args.style,
        path=args.path,
        clipboard=args.clipboard,
    )


if __name__ == '__main__':
    main()



from pygclip.io import run_shell_command, write_html_to_clipboard


def test_run_shell_command():
    """Test that a simple shell command returns the expected output."""
    assert run_shell_command(['echo', 'hello world']) == "hello world\n"


def test_write_html_to_clipboard(clipboard):
    """Test whether HTML is correctly encoded into the OSX clipboard."""
    write_html_to_clipboard('<p></p>')
    assert clipboard() == '<p></p>'
    assert clipboard(raw=True) == u"\xabdata HTML3C703E3C2F703E\xbb\n"



from io import StringIO
from subprocess import Popen, PIPE
import sys
import pytest

from pygclip.cli import main


def test_main_with_file(resource):
    """Test that the entry point will complete for a simple file."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()


def test_main_without_style_or_lexer(resource):
    """Test that the entry point accepts direct arguments not just `sys.argv`."""
    sys.argv = ['pygclip', resource('simple.py')]
    main()


def test_main_with_args(clipboard, resource):
    """Test that the entry point accepts direct arguments not just `sys.argv`."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()
    argv_result = clipboard()
    main(['-s', 'emacs', '-l', 'python', resource('simple.py')])
    args_result = clipboard()
    assert argv_result != args_result


def test_main_fails_with_nonexistent_file(resource):
    """Test that the entry point fails with a path to a file that doesn't exist."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('missing.py')]
    with pytest.raises(FileNotFoundError):
        main()


def test_main_with_file_debug(clipboard, resource):
    """Test that debug mode for the main entry point has no effect on output."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()
    normal_result = clipboard()

    sys.argv = ['pygclip', '-d', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()
    debug_result = clipboard()

    assert normal_result == debug_result


def test_main_with_stdin(clipboard, resource):
    """Test that the entry point will complete reading from standard input."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()
    file_result = clipboard()

    old_stdin = sys.stdin
    with open(resource('simple.py'), 'r') as f:
        sys.stdin = StringIO(f.read())
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python']
    main()
    stdin_result = clipboard()
    sys.stdin = old_stdin

    assert file_result == stdin_result


def test_main_with_clipboard(clipboard, resource):
    """Test that the entry point will complete reading from the clipboard."""
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', resource('simple.py')]
    main()
    file_result = clipboard()

    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', '-c']
    with open(resource('simple.py'), 'r') as f:
        text = f.read()
    Popen(['pbcopy'], stdin=PIPE).communicate(text.encode('utf-8'))
    main()
    clipboard_result = clipboard()

    assert file_result == clipboard_result


def test_main_fails_with_blank_clipboard():
    """Test that the entry point will fail when reading from a blank clipboard."""
    Popen(['osascript', '-e', 'set the clipboard to ""']).communicate()
    sys.argv = ['pygclip', '-s', 'monokai', '-l', 'python', '-c']
    with pytest.raises(ValueError):
        main()



"""
    pygclip.io
    ~~~~~~~~~~~

    Input/output functionality.

    :copyright: © 2018 by Bryan Lee McKelvey.
    :license: MIT, see LICENSE for more details.
"""

import logging
from subprocess import Popen, PIPE
from typing import List, Union


def run_shell_command(args: List[str], stdin: Union[bytes, str, None]=None) -> str:
    logger = logging.getLogger(__name__)
    logger.debug('Running command: {}'.format(args))
    if stdin is None:
        process = Popen(args, stdout=PIPE)
        return process.communicate()[0].decode('utf-8')
    else:
        process = Popen(args, stdin=PIPE, stdout=PIPE)
        return process.communicate(stdin.encode('utf-8'))[0].decode('utf-8')


def write_html_to_clipboard(html: str) -> None:
    hex_text = html.encode('utf-8').hex()
    run_shell_command(['osascript', '-e', u'set the clipboard to \xabdata HTML{}\xbb'.format(hex_text)])





from pygclip.logging import setup_logging


def test_setup_logging():
    """Test whether the logging level can be set to a default."""
    setup_logging()


def test_setup_logging_debug_mode():
    """Test whether the logging level can be set in debug mode."""
    setup_logging(True)




"""
    pygclip.logging
    ~~~~~~~~~~~~~~~

    Logging functions.

    :copyright: © 2018 by Bryan Lee McKelvey.
    :license: MIT, see LICENSE for more details.
"""

import logging


def setup_logging(debug: bool=False) -> None:
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)



import io
import re
import sys
from setuptools import setup

if sys.version_info < (3, 5):
    sys.exit('Sorry, Python < 3.5 is not supported.')

with io.open('README.rst', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('pygclip/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='pygclip',
    version=version,
    license='MIT',
    author='Bryan Lee McKelvey',
    author_email='bryan.mckelvey@gmail.com',
    url='https://github.com/brymck/pygclip',
    description='Pygmentize to clipboard for macOS.',
    long_description=readme,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=['pygclip'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'pygments',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
    ],
    extras_require={
        'dev': [
            'coverage',
            'pytest>=3',
            'sphinx',
            'tox',
        ],
    },
    entry_points={
        'console_scripts': [
            'pygclip = pygclip.cli:main',
        ],
    },
)



"""
    pygclip.html
    ~~~~~~~~~~~~~~

    HTML-related functions.

    :copyright: © 2018 by Bryan Lee McKelvey.
    :license: MIT, see LICENSE for more details.
"""

import logging
import re
import sys
from typing import Dict, Union
import xml.etree.ElementTree as ETree

from pygclip.colors import darken_color
from pygclip.io import run_shell_command


def _create_options_text(
    style: Union[str, None],
) -> str:
    options = ['noclasses']
    if style is not None:
        options.append('style=' + style)
    return ','.join(options)


def _get_styles(node: ETree.Element) -> Dict[str, str]:
    style_text = node.attrib.get('style', '')
    styles = {}
    for pair in style_text.split(';'):
        key, value = pair.split(':')
        styles[key.strip()] = value.strip()
    return styles


def _override_styles(node: ETree.Element, overrides: Dict[str, str]) -> None:
    styles = _get_styles(node)
    styles.update(overrides)
    pairs = ['{}: {}'.format(k, v) for k, v in styles.items()]
    style_text = '; '.join(pairs)
    node.set('style', style_text)


def _modify_html(html: str) -> str:
    root = ETree.fromstring(html)

    # Format <pre> tag within enclosing <div>
    background = _get_styles(root).get('background')
    border_color = darken_color(background, 0.5)
    _override_styles(root.find('pre'), {
        'background': background,
        'border': 'solid 1px {}'.format(border_color),
        'font-family': 'monospace',
        'padding': '1em',
    })

    return ETree.tostring(root, encoding='utf-8').decode('utf-8')


def generate_html(
    lexer: Union[str, None],
    style: Union[str, None],
    path: Union[str, None],
    clipboard: bool,
) -> str:
    logger = logging.getLogger(__name__)
    opts_text = _create_options_text(style)
    cmd = ['pygmentize', '-O', opts_text, '-f', 'html']
    if lexer:
        cmd.extend(['-l', lexer])
    if path is None:
        if clipboard:
            text = run_shell_command(['pbpaste', '-Prefer', 'txt'])
            if not text.strip():
                raise ValueError('No plain-text content in clipboard.')
            html = run_shell_command(cmd, stdin=text)
        else:
            html = run_shell_command(cmd, stdin=sys.stdin.read())
    else:
        cmd.append(path)
        html = run_shell_command(cmd)

    # Remove extraneous newlines at end of code block
    html = re.sub(r'\n+(?=</pre></div>$)', '', html)

    html = _modify_html(html)

    logger.debug(html)
    return html



"""
    pygclip.colors
    ~~~~~~~~~~~~~~

    Color-related functions.

    :copyright: © 2018 by Bryan Lee McKelvey.
    :license: MIT, see LICENSE for more details.
"""

from typing import Tuple

RGB = Tuple[int, int, int]


def _hex_to_rgb(hex_color: str) -> RGB:
    hex_color = hex_color.lstrip('#')
    red, green, blue = [int(hex_color[i:i+2], 16) for i in (0, 2, 4)]
    return red, green, blue


def _rgb_to_hex(rgb: RGB) -> str:
    return '#{:02x}{:02x}{:02x}'.format(*rgb)


def _darken_rgb(rgb: RGB, factor: float) -> RGB:
    red, green, blue = [int(x * (1 - factor)) for x in rgb]
    return red, green, blue


def darken_color(hex_color: str, factor: float) -> str:
    return _rgb_to_hex(_darken_rgb(_hex_to_rgb(hex_color), factor))

